from django.dispatch import Signal

__all__ = ['ipn_processed', 'state_changed']

ipn_processed = Signal(providing_args=['instance'])

state_changed = Signal(providing_args=['instance', 'previous_state'])



import importlib
from django.conf import settings


CACHE_BATCH_KEY_PREFIX = getattr(
    settings, 'WEPAY_CACHE_BATCH_KEY_PREFIX', 'wepay-cache-batch-key')

def make_batch_key(batch_id):
    return "%s-%s" % (CACHE_BATCH_KEY_PREFIX, batch_id)

def make_callback_key(batch_key, reference_id):
    return "%s-%s" % (batch_key, reference_id)

def from_string_import(string):
    """
    Returns the attribute from a module, specified by a string.
    """
    module, attrib = string.rsplit('.', 1)
    return getattr(importlib.import_module(module), attrib)



from django.contrib import admin

from djwepay.api import get_wepay_model, is_abstract
from djwepay.forms import AppForm

__all__ = ['AppAdmin', 'UserAdmin', 'AccountAdmin']

class AppAdmin(admin.ModelAdmin):
    list_display = ('client_id', 'status', 'production')
    list_filter = ('status', 'production', 'date_created')
    readonly_fields = ('status', 'date_created', 'date_modified')
    search_fields = ('status', 'theme_object', 'gaq_domain')

    form = AppForm

if not is_abstract('app'):
    admin.site.register(get_wepay_model('app'), AppAdmin)

class UserAdmin(admin.ModelAdmin):
    list_display = ('user_id', 'user_name', 'email', 'state')
    list_filter = ('state', 'date_created')
    readonly_fields = (
        'user_id', 'user_name', 'first_name', 'last_name', 'email', 'state', 
        'access_token', 'expires_in', 'date_created', 'date_modified'
    )
    search_fields = ('user_id', 'user_name', 'email', 'state')

if not is_abstract('user'):
    admin.site.register(get_wepay_model('user'), UserAdmin)

class AccountAdmin(admin.ModelAdmin):
    list_display = (
        'account_id', 'name', 'user', 'state', 'verification_state', 'type'
    )
    list_filter = ('state', 'verification_state', 'type', 'date_created')
    readonly_fields = (
        'account_id', 'user', 'state', 'payment_limit', 'verification_state', 
        'type', 'create_time', 'date_created', 'date_modified'
    )

if not is_abstract('account'):
    admin.site.register(get_wepay_model('account'), AccountAdmin)



"""All models are direct mappings to the WePay objects. By default only the
fields that correspond to the values returned from WePay lookup calls
(ex. `/account <https://www.wepay.com/developer/reference/account#lookup>`_) are
include in the models. All fields follow the rules outlined in `Storing Data
<https://www.wepay.com/developer/reference/storing_data>`_, unless otherwise
specified in object's documentation. For that reason values, which have there
names end with '_uri' (ex. ``account_uri``) are not included as model fields,
instead they are added as dynamic cached object properties, which are inherited
from Api objects defined in :mod:`djwepay.api`.

"""

import datetime, pytz
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db import models

from djwepay.api import *
from djwepay.fields import MoneyField

from json_field import JSONField

__all__ = ['App', 'User', 'Account', 'Checkout', 'Preapproval', 'Withdrawal', 
           'CreditCard']

APP_CACHE = {}

class BaseModel(models.Model):
    date_created  = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)

    def save(self, *args, **kwargs):
        ''' On save, update timestamps '''
        if not self.date_created:
            self.date_created = datetime.datetime.now(pytz.timezone(settings.TIME_ZONE))
            #datetime.datetime.today()
        self.date_modified = datetime.datetime.now(pytz.timezone(settings.TIME_ZONE))
        return super(BaseModel, self).save(*args, **kwargs)

    class Meta:
        abstract = True
        ordering = ['-date_created']

class AppManager(models.Manager):

    def get_current(self):
        """Returns the current :class:`App` based on the :ref:`WEPAY_APP_ID` in the
        project's settings. The :class:`App` object is cached the first time
        it's retrieved from the database.

        """
        try:
            app_id = settings.WEPAY_APP_ID
        except AttributeError:
            raise ImproperlyConfigured(
                "You're using the Django WePay application without having set the "
                "WEPAY_APP_ID setting. Create an app in your database and set the "
                "WEPAY_APP_ID setting to fix this error.")
        try:
            current_app = APP_CACHE[app_id]
        except KeyError:
            current_app = self.get(pk=app_id)
            APP_CACHE[app_id] = current_app
        return current_app

    def clear_cache(self):
        """Clears the ``App`` object cache."""
        global APP_CACHE
        APP_CACHE = {}

class App(AppApi, BaseModel):
    """
    This model stores all of the relevant WePay application information. Only one
    instance of it at a time is supported per django application, which is 
    controlled by :ref:`WEPAY_APP_ID` setting.
    """
    client_id = models.BigIntegerField(primary_key=True)
    status = models.CharField(max_length=255)
    theme_object = JSONField(null=True, blank=True)
    gaq_domains = JSONField(null=True, blank=True)

    client_secret = models.CharField(max_length=255)
    access_token = models.CharField(max_length=255)
    production = models.BooleanField(default=True)
    state = models.CharField(max_length=255)

    objects = AppManager()

    class Meta(BaseModel.Meta):
        abstract = is_abstract('app')
        db_table = 'djwepay_app'
        verbose_name = 'WePay App'

class UserManager(models.Manager):

    def create_from_response(self, response):
        try:
            user = self.get(pk=response['user_id'])
        except self.model.DoesNotExist:
            user = self.model()
        return user.instance_update(response)
    
    def accessible(self):
        return self.exclude(access_token=None)

class User(UserApi, BaseModel):
    user_id = models.BigIntegerField(primary_key=True)
    user_name = models.CharField(max_length=255)
    first_name = models.CharField(max_length=255)
    last_name = models.CharField(max_length=255)
    email = models.EmailField(max_length=255)
    state = models.CharField(max_length=255)

    # access_token=NULL means it has been revoked.
    access_token = models.CharField(null=True, max_length=255)
    token_type = "BEARER"
    expires_in = models.BigIntegerField(null=True, blank=True)

    objects = UserManager()

    def __str__(self):
        return self.user_name

    class Meta(BaseModel.Meta):
        abstract = is_abstract('user')
        db_table = 'djwepay_user'
        verbose_name = 'WePay User'

class AccountManager(models.Manager):
    
    def create_from_response(self, user, response):
        account = self.model(user=user)
        return account.instance_update(response)

    def accessible(self):
        return self.exclude(user__access_token=None)

    def active(self):
        return self.accessible().filter(state='active')


class Account(AccountApi, BaseModel):
    account_id = models.BigIntegerField(primary_key=True)
    user = models.ForeignKey(get_wepay_model_name('user'), related_name='accounts')
    name = models.CharField(max_length=255)
    state = models.CharField(max_length=255)
    description = models.CharField(max_length=255)
    reference_id = models.CharField(max_length=255)
    gaq_domains = JSONField(null=True, blank=True)
    theme_object = JSONField(null=True, blank=True)
    type = models.CharField(max_length=255)
    create_time = models.BigIntegerField(null=True)
    balances = JSONField(null=True, blank=True)
    statuses = JSONField(null=True, blank=True)
    action_reasons = JSONField(null=True, blank=True)
    country = models.CharField(max_length=2)
    currencies = JSONField(null=True, blank=True)

    objects = AccountManager()
    
    def __str__(self):
        return "%s - %s" % (self.pk, self.name)

    class Meta(BaseModel.Meta):
        abstract = is_abstract('account')
        db_table = 'djwepay_account'
        verbose_name = 'WePay Account'

class AccountObjectsManager(models.Manager):
    
    def create_from_response(self, account, response):
        obj = self.model(account=account)
        return obj.instance_update(response)

    def accessible(self):
        return self.exclude(account__user__access_token=None)


class Checkout(CheckoutApi, BaseModel):
    checkout_id = models.BigIntegerField(primary_key=True)
    account = models.ForeignKey(
        get_wepay_model_name('account'), related_name='checkouts')
    preapproval = models.ForeignKey(
        get_wepay_model_name('preapproval'), related_name='checkouts', null=True)
    state = models.CharField(max_length=255)
    soft_descriptor = models.CharField(max_length=255)
    short_description = models.CharField(max_length=255)
    long_description = models.CharField(max_length=2047, blank=True)
    currency = "USD"
    amount = MoneyField(null=True)
    fee = MoneyField(null=True)
    gross = MoneyField(null=True)
    app_fee = MoneyField(null=True)
    fee_payer = models.CharField(max_length=255)
    reference_id = models.CharField(max_length=255, blank=True)
    payer_email = models.EmailField(max_length=255, blank=True)
    payer_name = models.CharField(max_length=255, blank=True)
    cancel_reason = models.CharField(max_length=255, blank=True)
    refund_reason = models.CharField(max_length=255, blank=True)
    auto_capture = models.BooleanField(default=True)
    require_shipping = models.BooleanField(default=False)
    shipping_address = JSONField(null=True)
    tax = MoneyField(null=True)
    amount_refunded = MoneyField(null=True)
    amount_charged_back = MoneyField(null=True)
    create_time = models.BigIntegerField(null=True)
    mode = models.CharField(max_length=255)

    objects = AccountObjectsManager()

    def __str__(self):
        return "%s - %s" % (self.pk, self.short_description)

    class Meta(BaseModel.Meta):
        abstract = is_abstract('checkout')
        db_table = 'djwepay_checkout'
        verbose_name = 'WePay Checkout'


class Preapproval(PreapprovalApi, BaseModel):
    preapproval_id = models.BigIntegerField(primary_key=True)
    account = models.ForeignKey(
        get_wepay_model_name('account'), related_name='preapprovals')
    short_description = models.CharField(max_length=255)
    long_description = models.CharField(max_length=2047, blank=True)
    currency = "USD"
    amount = MoneyField(null=True)
    fee_payer = models.CharField(max_length=255)
    state = models.CharField(max_length=255)
    app_fee = MoneyField(null=True)
    period = models.CharField(max_length=255)
    frequency = models.IntegerField(null=True)
    start_time = models.BigIntegerField(null=True)
    end_time = models.BigIntegerField(null=True)
    reference_id = models.CharField(max_length=255)
    shipping_address = JSONField(null=True)
    shipping_fee = MoneyField(null=True)
    tax = MoneyField(null=True)
    auto_recur = models.BooleanField(default=False)
    payer_name = models.CharField(max_length=255)
    payer_email = models.EmailField(max_length=255, blank=True)
    create_time = models.BigIntegerField(null=True)
    next_due_time = models.BigIntegerField(null=True)
    last_checkout = models.ForeignKey(
        get_wepay_model_name('checkout'), null=True, related_name='+')
    last_checkout_time = models.BigIntegerField(null=True)
    mode = models.CharField(max_length=255)

    objects = AccountObjectsManager()

    def __str__(self):
        return "%s - %s" % (self.pk, self.short_description)

    class Meta(BaseModel.Meta):
        abstract = is_abstract('preapproval')
        db_table = 'djwepay_preapproval'
        verbose_name = 'WePay Preapproval'


class Withdrawal(WithdrawalApi, BaseModel):
    withdrawal_id = models.BigIntegerField(primary_key=True)
    account = models.ForeignKey(
        get_wepay_model_name('account'), related_name='withdrawals')
    state = models.CharField(max_length=255)
    amount = MoneyField(null=True)
    note = models.CharField(max_length=255)
    recipient_confirmed = models.NullBooleanField()
    type = models.CharField(max_length=255)
    create_time = models.BigIntegerField(null=True)

    objects = AccountObjectsManager()

    def __str__(self):
        return "%s - %s" % (self.pk, self.amount)

    class Meta(BaseModel.Meta):
        abstract = is_abstract('withdrawal')
        db_table = 'djwepay_withdrawal'
        verbose_name = 'WePay Preapproval'


class CreditCard(CreditCardApi, BaseModel):
    credit_card_id = models.BigIntegerField(primary_key=True)
    credit_card_name = models.CharField(max_length=255)
    state = models.CharField(max_length=255)
    user_name = models.CharField(max_length=255)
    email = models.CharField(max_length=255, blank=True)
    reference_id = models.CharField(max_length=255, blank=True)
    create_time = models.BigIntegerField(null=True)

    def __str__(self):
        return self.credit_card_name

    class Meta(BaseModel.Meta):
        abstract = is_abstract('credit_card')
        db_table = 'djwepay_credit_card'
        verbose_name = 'WePay Credit Card'


class SubscriptionPlan(SubscriptionPlanApi, BaseModel):
    subscription_plan_id = models.BigIntegerField(primary_key=True)
    account = models.ForeignKey(
        get_wepay_model_name('account'), related_name='subscription_plans')
    name = models.CharField(max_length=255)
    short_description = models.CharField(max_length=2047)
    currency = models.CharField(max_length=3)
    amount = MoneyField(null=True)
    period = models.CharField(max_length=255)
    app_fee = MoneyField(null=True)
    fee_payer = models.CharField(max_length=255)
    state = models.CharField(max_length=255)
    create_time = models.BigIntegerField(null=True)
    number_of_subscriptions = models.BigIntegerField(null=True)
    trial_length = models.BigIntegerField(null=True)
    setup_fee = MoneyField(null=True)
    reference_id = models.CharField(max_length=255)
    
    objects = AccountObjectsManager()

    class Meta(BaseModel.Meta):
        abstract = is_abstract('subscription_plan')
        db_table = 'djwepay_subscription_plan'
        verbose_name = 'WePay Subscription Plan'


class SubscriptionManager(models.Manager):
    
    def create_from_response(self, subscription_plan, response):
        obj = self.model(subscription_plan=subscription_plan)
        return obj.instance_update(response)

    def accessible(self):
        return self.exclude(subscription_plan__account__user__access_token=None)


class Subscription(SubscriptionApi, BaseModel):
    subscription_id = models.BigIntegerField(primary_key=True)
    subscription_plan = models.ForeignKey(
        get_wepay_model_name('subscription_plan'), related_name='subscriptions')
    payer_name = models.CharField(max_length=255)
    payer_email = models.CharField(max_length=255)
    currency = models.CharField(max_length=255)
    amount = MoneyField(null=True)
    period = models.CharField(max_length=255)
    app_fee = MoneyField(null=True)
    fee_payer = models.CharField(max_length=255)
    state = models.CharField(max_length=255)
    create_time = models.BigIntegerField(null=True)
    payment_method_id = models.BigIntegerField(null=True)
    payment_method_type = models.CharField(max_length=255)
    quantity = models.BigIntegerField(null=True)
    mode = models.CharField(max_length=255)
    trial_days_remaining = models.BigIntegerField(null=True)
    transition_expire_time = models.BigIntegerField(null=True)
    transition_prorate = models.NullBooleanField()
    transition_quantity = models.BigIntegerField(null=True)
    transition_subscription_plan = models.ForeignKey(
        get_wepay_model_name('subscription_plan'), 
        related_name='transition_subscriptions')
    reference_id = models.CharField(max_length=255)

    objects = SubscriptionManager()

    class Meta(BaseModel.Meta):
        abstract = is_abstract('subscription')
        db_table = 'djwepay_subscription'
        verbose_name = 'WePay Subscription'


class SubscriptionCharge(SubscriptionChargeApi, BaseModel):
    subscription_charge_id = models.BigIntegerField(primary_key=True)
    subscription_plan = models.ForeignKey(
        get_wepay_model_name('subscription_plan'), related_name='subscription_charges')
    subscription = models.ForeignKey(
        get_wepay_model_name('subscription'), related_name='subscription_charges')
    type = models.CharField(max_length=255)
    amount = MoneyField(null=True)
    currency = models.CharField(max_length=3)
    fee = MoneyField(null=True)
    app_fee = MoneyField(null=True)
    gross = MoneyField(null=True)
    quantity = models.BigIntegerField(null=True)
    amount_refunded = MoneyField(null=True)
    amount_charged_back = MoneyField(null=True)
    state = models.CharField(max_length=255)
    create_time = models.BigIntegerField(null=True)
    end_time = models.BigIntegerField(null=True)
    prorate_time = models.BigIntegerField(null=True)

    class Meta(BaseModel.Meta):
        abstract = is_abstract('subscription_charge')
        db_table = 'djwepay_subscription_charge'
        verbose_name = 'WePay Subscription Charge'



from django.http import Http404, HttpResponse
from django.shortcuts import get_object_or_404
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic.base import View

from djwepay.models import User
from djwepay.signals import ipn_processed
from djwepay.api import get_wepay_model
from wepay.exceptions import WePayError

__all__ = ['IPNView', 'OAuth2Mixin', 'TestsCallbackView']

class IPNView(View):
    http_method_names = ['post']

    @method_decorator(csrf_exempt)
    def dispatch(self, *args, **kwargs):
        return super(IPNView, self).dispatch(*args, **kwargs)

    def post(self, request, obj_name=None, user_id=None, **kwargs):
        if obj_name == 'preapproval' and 'checkout_id' in request.POST:
            # as described in https://www.wepay.com/developer/reference/ipn
            # checkouts created automatically will receive IPNs to preapproval's
            # callback_uri
            obj_name = 'checkout'
        model = get_wepay_model(obj_name)
        obj_id_name = "%s_id" % obj_name
        user = None
        try:
            obj_id = request.POST[obj_id_name]
        except KeyError:
            return HttpResponse(
                "Missing object_id in POST: %s" % obj_id_name, status=501)
        try:
            obj = model.objects.get(pk=obj_id)
        except model.DoesNotExist:
            if obj_name == 'user':
                raise Http404("User object with user_id: '%s' not found." % obj_id)
            obj = model(pk=obj_id)
        if not user_id is None:
            user = get_object_or_404(get_wepay_model('user'), pk=user_id)
            obj.access_token = user.access_token
        try:
            api_call = getattr(obj, "api_%s" % obj_name)
            api_call()
        except WePayError, e:
            if e.code == 1011 and not user is None: # acess_token has been revoked
                user.access_token = None
                user.save()
            else:
                return HttpResponse(
                    "WePay error on update. %s" % e, status=500)
        ipn_processed.send(sender=model, instance=obj)
        return HttpResponse("Successfull %s update." % obj_name)
        

class OAuth2Mixin(object):
    """
    Mixin for Django generic style views that is helpfull for WePay OAuth2 process
    """

    @property
    def app(self):
        return get_wepay_model('app').objects.get_current()
    
    def get_redirect_uri(self):
        """
        Returns current view's url. Override this method to supply a different
        redirect_uri for oauth2 calls.
        """
        return self.request.path

    def get_authorization_url(self, user=None, prefill=True, **kwargs):
        """Calls :meth:`djwepay.api.AppApi.api_oauth2_authorize` and returns a url where
        user can be sent off to in order to grand access. Prefills ``redirect_uri``
        by calling :func:`get_redirect_uri`. ``user_name`` and ``user_email`` 
        parameters are prefilled by using information from django :attr:`user`
        object, which can be passed as a keyword argument otherwise it defaults to 
        ``self.request.user``, in order to prevent that behavior pass ``prefill=False``. 
        Any extra keywords will be passed along to the api call.

        :keyword User user: Will be used to prefill ``user_name`` and
            ``user_email``. Defaults to ``self.request.user``.

        :keyword bool prefill: set to ``False`` in order to prevent user info retrival.

        """

        if prefill:
            user = user or self.request.user
            if not user.is_anonymous():
                if not 'user_name' in kwargs:
                    kwargs['user_name'] = user.get_full_name()
                if not 'user_email' in kwargs:
                    kwargs['user_email'] = user.email
        return self.app.api_oauth2_authorize(
            redirect_uri=self.get_redirect_uri(), **kwargs)

    def get_user(self, **kwargs):
        """
        Calls :func:`djwepay.api.App.api_oauth2_token`
        """
        if 'error' in self.request.GET and \
           self.request.GET['error'] == "access_denied":
            raise AttributeError("%s - %s" % (
                self.request.GET['error'], 
                self.request.GET.get('error_description', '')))
        try:
            code = self.request.GET['code']
        except KeyError:
            raise AttributeError("'code' is missing from GET parameters")
        try:
            user, response = self.app.api_oauth2_token(
                code=code, redirect_uri=self.get_redirect_uri(), **kwargs)
            return user
        except WePayError, e:
            if e.code == 1012: # the code has expired
                raise AttributeError(str(e))
            raise



class TestsCallbackView(View):
    """
    This view is used whenever the tests for djwepay app are being run.
    All the api calls in the test suite run in a test enviroment (test db, 
    no webserver etc.), but since the calls are made to the actual 
    stage.wepay.com server we cannot give real callback_uri's instead we use this
    fake view to send a 200 response in order to prevent IPN retries.
    """
    http_method_names = ['post', 'get']

    @method_decorator(csrf_exempt)
    def dispatch(self, *args, **kwargs):
        return super(TestsCallbackView, self).dispatch(*args, **kwargs)

    def post(self, request, **kwargs):
        return HttpResponse("Success.")

    def get(self, request, **kwargs):
        return HttpResponse("Success.")




from django.core.cache import cache
from django.conf import settings

__all__ = ['cached_property']


class cached_property(object):
    def __init__(self, fget):
        self.fget = fget
        self.__doc__ = fget.__doc__
        self.__name__ = fget.__name__
        self.__module__ = fget.__module__
        self._key = "_%s" % fget.__name__

    def __get__(self, obj, cls):
        if obj is None:
            return self
        value = getattr(obj, self._key, None)
        if value is None:
            value = self.fget(obj)
            setattr(obj, self._key, value)
        return value

    def __set__(self, obj, value):
        setattr(obj, self._key, value)



from django import forms

from djwepay.api import get_wepay_model

class AppForm(forms.ModelForm):
    def save(self, *args, **kwargs):
        if self.instance:
            self.instance.api_app()
        return super(AppForm, self).save(*args, **kwargs)

    class Meta:
        model = get_wepay_model('app')

ACCOUNT_TYPE_CHOICES = (
    ('personal', u"Personal"),
    ('nonprofit', u"Non-profit Organization"),
    ('business', u"Business"),
)

class AccountCreateForm(forms.Form):
    name = forms.CharField(max_length=255, required=True,
                           help_text=u"The name of the account you want to create.")
    description = forms.CharField(
        max_length=255, required=True, help_text=u"The description of the account "
        "you want to create.", widget=forms.Textarea)
    type = forms.ChoiceField(required=False, choices=ACCOUNT_TYPE_CHOICES)

    def clean_name(self):
        name = self.cleaned_data['name']
        if name.lower().rfind("wepay") >= 0:
            raise forms.ValidationError("Account name cannot contain 'wepay'.")
        return name



from django.conf.urls import patterns, include, url

from djwepay.models import *
from djwepay.views import *


urlpatterns = patterns('',
    url(r'^ipn/(?P<obj_name>user|account|checkout|preapproval|withdrawal|'
        'subscription_plan|subscription)/(?:(?P<user_id>\d+)/)?$', 
        IPNView.as_view(), name="ipn"),
    # used for tests
    url(r'^tests_callback/$', TestsCallbackView.as_view(), name="tests_callback"),
)



from django.conf import settings
from django.db import models

# Add MoneyField to 'south', so migrations are possible
try:
    if 'south' in getattr(settings, 'INSTALLED_APPS'):
        from south.modelsinspector import add_introspection_rules
        add_introspection_rules([], ["^djwepay\.fields\.MoneyField"])
except ImportError: pass

class MoneyField(models.DecimalField):
    def __init__(self, *args, **kwargs):
        if not 'decimal_places' in kwargs:
            kwargs['decimal_places'] = 2
        if not 'max_digits' in kwargs:
            kwargs['max_digits'] = 11 # ~1 billion USD, should be enough for now
        super(MoneyField, self).__init__(*args, **kwargs)




from django.conf import settings
from django.core.cache import cache
from django.core.urlresolvers import reverse
from django.db.models.loading import get_model
from django.utils.functional import LazyObject, curry

from djwepay.decorators import cached_property
from djwepay.signals import state_changed
from djwepay.utils import from_string_import
from wepay.exceptions import WePayError

__all__ = ['AppApi', 'UserApi', 'AccountApi', 'CheckoutApi', 'PreapprovalApi',
           'WithdrawalApi', 'CreditCardApi', 'SubscriptionPlanApi',
           'SubscriptionApi', 'SubscriptionChargeApi', 'DEFAULT_SCOPE',
           'get_wepay_model', 'get_wepay_model_name', 'is_abstract']

# default is full access
DEFAULT_SCOPE = getattr(
    settings, 'WEPAY_DEFAULT_SCOPE', "manage_accounts,collect_payments,"
    "view_balance,view_user,preapprove_payments,send_money")

DEFAULT_MODELS = (
    ('app', 'djwepay.App'),
    ('user', 'djwepay.User'),
    ('account', 'djwepay.Account'),
    ('checkout', 'djwepay.Checkout'),
    ('preapproval', 'djwepay.Preapproval'),
    ('withdrawal', 'djwepay.Withdrawal'),
    ('credit_card', 'djwepay.CreditCard'),
    ('subscription_plan', 'djwepay.SubscriptionPlan'),
    ('subscription', 'djwepay.Subscription'),
    ('subscription_charge', 'djwepay.SubscriptionCharge')
)

MODELS = getattr(settings, 'WEPAY_MODELS', ())

API_BACKEND = getattr(settings, 'WEPAY_API_BACKEND', 'djwepay.backends.default.WePay')

def get_wepay_model_name(obj_name):
    return dict(DEFAULT_MODELS + MODELS).get(obj_name)

def get_wepay_model(obj_name):
    model_name = get_wepay_model_name(obj_name)
    if model_name is None:
        return None
    return get_model(*model_name.split('.'))

def is_abstract(obj_name):
    return not get_wepay_model_name(obj_name) == dict(DEFAULT_MODELS).get(obj_name)
    

class WePayLazy(LazyObject):
    def _setup(self):
        backend = from_string_import(API_BACKEND)
        app = get_wepay_model('app').objects.get_current()
        self._wrapped = backend(
            production=app.production, access_token=app.access_token, timeout=45)

class AppLazy(LazyObject):
    def _setup(self):
        self._wrapped = get_wepay_model('app').objects.get_current()


class Api(object):

    api = WePayLazy()
    app = AppLazy()

    def instance_update(self, response):
        previous_state = getattr(self, 'state', '') # app object doesn't have state?
        new_state = response.get('state', '')
        for key, value in response.iteritems():
            setattr(self, key, value)
        if new_state and new_state != previous_state:
            # using cache we eliminate duplicate calls to state_changed,
            # which has a chance of happening in multithreaded environment
            cache_key = "wepay-state-changed-%s-%s" % (type(self).__name__, self.pk)
            added = cache.add(cache_key, new_state)
            if not added:
                stored_state = cache.get(cache_key)
                if stored_state == new_state:
                    return response
            cache.set(cache_key, new_state)                
            state_changed.send(sender=type(self), instance=self, 
                               previous_state=previous_state)
        self.save()
        return self

    def instance_identity(self, response):
        return self

    def get_callback_uri(self, **kwargs):
        return reverse('wepay:ipn', kwargs=kwargs)
    
    def api_batch_create(self, **kwargs):
        return self.api.batch.create(client_id=self.app.client_id,
                                     client_secret=self.app.client_secret, **kwargs)

class AppApi(Api):
    """ App model mixin object that helps making related Api calls"""

    def api_app(self, **kwargs):
        return self.api.app(
            client_id=self.client_id,
            client_secret=self.client_secret, 
            callback=self.instance_update, **kwargs)

    def api_app_modify(self, **kwargs):
        return self.api.app.modify(
            client_id=self.client_id,
            client_secret=self.client_secret, 
            callback=self.instance_update, **kwargs)

    def api_oauth2_authorize(self, redirect_uri, **kwargs):
        return self.api.oauth2.authorize(
            self.client_id, redirect_uri, DEFAULT_SCOPE, **kwargs)

    def api_oauth2_token(self, **kwargs):
        User = get_wepay_model('user')
        return self.api.oauth2.token(
            client_id=self.client_id,
            client_secret=self.client_secret,
            callback_uri=self.get_callback_uri(obj_name='user'),
            callback=User.objects.create_from_response, **kwargs)

    def api_user_register(self, **kwargs):
        User = get_wepay_model('user')
        return self.api.user.register(
            client_id=self.client_id,
            client_secret=self.client_secret,
            scope=DEFAULT_SCOPE,
            callback_uri=self.get_callback_uri(obj_name='user'),
            callback=User.objects.create_from_response, **kwargs)


    def api_preapproval_create(self, **kwargs):
        Preapproval = get_wepay_model('preapproval')
        return self.api.preapproval.create(
            client_id=self.client_id, 
            client_secret=self.client_secret,
            callback_uri = self.get_callback_uri(obj_name='preapproval'),
            callback=curry(Preapproval.objects.create_from_response, None), 
            **kwargs)

    def api_preapproval_find(self, **kwargs):
        return self.api.preapproval.find(**kwargs)

    def api_credit_card_create(self, **kwargs):
        CreditCard = get_wepay_model('credit_card')
        return self.api.credit_card.create(
            client_id=self.client_id, 
            callback=CreditCard.objects.create_from_response, **kwargs)

    def api_credit_card_find(self, **kwargs):
        return self.api.credit_card.find(
            client_id=self.client_id,
            client_secret=self.client_secret, **kwargs)

    def api_subscription_plan_find(self, **kwargs):
        return self.api.subscription_plan.find(
            account_id=self.pk,
            access_token=self.access_token, **kwargs)


class UserApi(Api):

    def api_user(self, **kwargs):
        return self.api.user(
            access_token=self.access_token, 
            callback=self.instance_update, **kwargs)

    def api_user_modify(self, **kwargs):
        return self.api.user.modify(
            access_token=self.access_token,
            callback=self.instance_update, **kwargs)

    def api_user_resend_confirmation(self, **kwargs):
        return self.api.user.resend_confirmation(
            access_token=self.access_token,
            callback=self.instance_update, **kwargs)

    def api_account_create(self, **kwargs):
        Account = get_wepay_model('account')
        return self.api.account.create(
            access_token=self.access_token,
            callback_uri=self.get_callback_uri(
                obj_name='account', user_id=self.user_id),
            callback=curry(Account.objects.create_from_response, self), **kwargs)

    def api_account_find(self, **kwargs):
        return self.api.account.find(
            access_token=self.access_token, **kwargs)

class AccountApi(Api):

    @cached_property
    def access_token(self):
        return self.user.access_token

    @cached_property
    def uri(self):
        return self.api_get_update_uri()[1].get('uri', None)

    def api_account(self, **kwargs):
        try:
            return self.api.account(
                account_id=self.pk, 
                access_token=self.access_token, 
                callback=self.instance_update, **kwargs)
        except WePayError, e:
            if e.code == 3003: # The account has been deleted
                self.state = 'deleted'
                self.save()
            raise

    def api_account_modify(self, **kwargs):
        return self.api.account.modify(
            account_id=self.pk, 
            access_token=self.access_token, 
            callback=self.instance_update, **kwargs)
        
    def api_account_delete(self, **kwargs):
        return self.api.account.delete(
            account_id=self.pk, 
            access_token=self.access_token, 
            callback=self.instance_update, **kwargs)

    def api_account_get_update_uri(self, **kwargs):
        instance, response = self.api.account.get_update_uri(
            account_id=self.pk, 
            access_token=self.access_token, 
            callback=self.instance_identity, **kwargs)
        self.uri = response['uri']
        return (instance, response)
        
    def api_account_get_reserve_details(self, **kwargs):
        return self.api.account.getreserve_details(
            account_id=self.pk, 
            access_token=self.access_token, 
            callback=self.instance_identity, **kwargs)

    def api_checkout_create(self, **kwargs):
        Checkout = get_wepay_model('checkout')
        return self.api.checkout.create(
            account_id=self.pk,
            access_token=self.access_token,
            callback_uri=self.get_callback_uri(
                obj_name='checkout', user_id=self.user.pk), 
            callback=curry(Checkout.objects.create_from_response, self), 
            **kwargs)

    def api_preapproval_create(self, **kwargs):
        Preapproval = get_wepay_model('preapproval')
        return self.api.preapproval.create(
            account_id=self.pk,
            access_token=self.access_token,
            callback_uri=self.get_callback_uri(
                obj_name='preapproval', user_id=self.user.pk), 
            callback=curry(Preapproval.objects.create_from_response, self), 
            **kwargs)

    def api_withdrawal_create(self, **kwargs):
        Withdrawal = get_wepay_model('withdrawal')
        return self.api.withdrawal.create(
            account_id=self.pk,
            access_token=self.access_token,
            callback_uri=self.get_callback_uri(
                obj_name='withdrawal', user_id=self.user.pk), 
            callback=curry(Withdrawal.objects.create_from_response, self), 
            **kwargs)

    def api_subscription_plan_create(self, **kwargs):
        SubscriptionPlan = get_wepay_model('subscription_plan')
        return self.api.subscription_plan.create(
            account_id=self.pk,
            access_token=self.access_token,
            callback_uri=self.get_callback_uri(
                obj_name='subscription_plan', user_id=self.user.pk), 
            callback=curry(SubscriptionPlan.objects.create_from_response, self), 
            **kwargs)

    def api_checkout_find(self, **kwargs):
        return self.api.checkout.find(
            account_id=self.pk, 
            access_token=self.access_token, **kwargs)

    def api_preapproval_find(self, **kwargs):
        return self.api.preapproval.find(
            account_id=self.account_id, 
            access_token=self.access_token, **kwargs)

    def api_withdrawal_find(self, **kwargs):
        return self.api.withdrawal.find(
            account_id=self.pk, 
            access_token=self.access_token, **kwargs)

    def api_subscription_plan_find(self, **kwargs):
        return self.api.subscription_plan.find(
            account_id=self.pk,
            access_token=self.access_token, **kwargs)

    def api_subscription_plan_get_button(self, **kwargs):
        return self.api.subscription_plan.get_button(
            account_id=self.pk,
            access_token=self.access_token, 
            callback=self.instance_identity, **kwargs)


class CheckoutApi(Api):

    @cached_property
    def access_token(self):
        return self.account.access_token

    @cached_property
    def redirect_uri(self):
        return self.api_checkout()[1].get('redirect_uri', None)

    @cached_property
    def callback_uri(self):
        return self.api_checkout()[1].get('callback_uri', None)

    @cached_property
    def dispute_uri(self):
        return self.api_checkout()[1].get('dispute_uri', None)
        
    def api_checkout(self, **kwargs):
        return self.api.checkout(
            checkout_id=self.pk, 
            access_token=self.access_token, 
            callback=self.instance_update, **kwargs)

    def api_checkout_cancel(self, **kwargs):
        return self.api.checkout.cancel(
            checkout_id=self.pk, 
            access_token=self.access_token, 
            callback=self.instance_update, **kwargs)

    def api_checkout_refund(self, **kwargs):
        return self.api.checkout.refund(
            checkout_id=self.pk, 
            access_token=self.access_token, 
            callback=self.instance_update, **kwargs)

    def api_checkout_capture(self, **kwargs):
        return self.api.checkout.capture(
            checkout_id=self.pk, 
            access_token=self.access_token, 
            callback=self.instance_update, **kwargs)

    def api_checkout_modify(self, **kwargs):
        return self.api.checkout.modify(
            checkout_id=self.pk, 
            access_token=self.access_token, 
            callback=self.instance_update, **kwargs)


class PreapprovalApi(Api):

    @cached_property
    def access_token(self):
        return self.account.access_token

    @cached_property
    def preapproval_uri(self):
        return self.api_preapproval()[1].get('preapproval_uri', None)

    @cached_property
    def manage_uri(self):
        return self.api_preapproval()[1].get('manage_uri', None)

    @cached_property
    def redirect_uri(self):
        return self.api_preapproval()[1].get('redirect_uri', None)

    @cached_property
    def callback_uri(self):
        return self.api_preapproval()[1].get('callback_uri', None)


    def api_preapproval(self, **kwargs):
        return self.api.preapproval(
            preapproval_id=self.pk, 
            access_token=self.access_token, 
            callback=self.instance_update, **kwargs)

    def api_preapproval_cancel(self, **kwargs):
        return self.api.preapproval.cancel(
            preapproval_id=self.pk, 
            access_token=self.access_token, 
            callback=self.instance_update, **kwargs)

    def api_preapproval_modify(self, **kwargs):
        return self.api.preapproval.modify(
            preapproval_id=self.pk, 
            access_token=self.access_token, 
            callback=self.instance_update, **kwargs)

    def api_checkout_create(self, **kwargs):
        Checkout = get_wepay_model('checkout')
        return self.api.checkout.create(
            account_id=self.account.pk,
            access_token=self.access_token,
            callback_uri=self.get_callback_uri(
                obj_name='checkout', user_id=self.user.pk), 
            callback=curry(Checkout.objects.create_from_response, self.account), 
            **kwargs)


class WithdrawalApi(Api):

    @cached_property
    def access_token(self):
        return self.account.access_token

    @cached_property
    def withdrawal_uri(self):
        return self.api_withdrawal()[1].get('withdrawal_uri', None)

    @cached_property
    def redirect_uri(self):
        return self.api_withdrawal()[1].get('redirect_uri', None)

    @cached_property
    def callback_uri(self):
        return self.api_withdrawal()[1].get('callback_uri', None)


    def api_withdrawal(self, **kwargs):
        return self.api.withdrawal(
            withdrawal_id=self.pk, 
            access_token=self.access_token, 
            callback=self.instance_update, **kwargs)

    def api_withdrawal_modify(self, **kwargs):
        return self.api.withdrawal.modify(
            withdrawal_id=self.pk, 
            access_token=self.access_token, 
            callback=self.instance_update, **kwargs)
        

class CreditCardApi(Api):
    def api_credit_card(self, **kwargs):
        return self.api.credit_card(
            client_id=self.app.client_id,
            client_secret=self.app.client_secret,
            credit_card_id=self.pk, 
            callback=self.instance_update, **kwargs)

    def api_credit_card_authorize(self, **kwargs):
        return self.api.credit_card.authorize(
            client_id=self.app.client_id,
            client_secret=self.app.client_secret,
            credit_card_id=self.pk, 
            callback=self.instance_update, **kwargs)

    def api_credit_card_delete(self, **kwargs):
        return self.api.credit_card.delete(
            client_id=self.app.client_id,
            client_secret=self.app.client_secret,
            credit_card_id=self.pk, 
            callback=self.instance_update, **kwargs)


class SubscriptionPlanApi(Api):

    @cached_property
    def access_token(self):
        return self.account.access_token
    
    @cached_property
    def callback_uri(self):
        return self.api_subscription_plan()[1].get('callback_uri', None)

    def api_subscription_plan(self, **kwargs):
        return self.api.subscription_plan(
            subscription_plan_id=self.pk,
            access_token=self.access_token, 
            callback=self.instance_update, **kwargs)

    def api_subscription_plan_delete(self, **kwargs):
        return self.api.subscription_plan.delete(
            subscription_plan_id=self.pk,
            access_token=self.access_token, 
            callback=self.instance_update, **kwargs)
            
    def api_subscription_plan_get_button(self, **kwargs):
        return self.api.subscription_plan.get_button(
            account_id=self.account.pk,
            button_type='subscription_plan',
            subscription_plan_id=self.pk,
            access_token=self.access_token, 
            callback=self.instance_identity, **kwargs)

    def api_subscription_plan_modify(self, **kwargs):
        return self.api.subscription_plan.modify(
            subscription_plan_id=self.pk,
            access_token=self.access_token, 
            callback=self.instance_update, **kwargs)

    def api_subscription_create(self, **kwargs):
        Subscription = get_wepay_model('subscription')
        return self.api.subscription.create(
            subscription_plan_id=self.pk,
            callback_uri=self.get_callback_uri(
                obj_name='subscription', user_id=self.account.user.pk),
            callback=curry(Subscription.objects.create_from_response, self),
            access_token=self.access_token, **kwargs)

    def api_subscription_find(self, **kwargs):
        return self.api.subscription.find(
            subscription_plan_id=self.pk,
            access_token=self.access_token, **kwargs)


class SubscriptionApi(Api):

    @cached_property
    def access_token(self):
        return self.subscription_plan.access_token
    
    @cached_property
    def subscription_uri(self):
        return self.api_subscription()[1].get('subscription_uri', None)

    @cached_property
    def callback_uri(self):
        return self.api_subscription()[1].get('callback_uri', None)

    @cached_property
    def redirect_uri(self):
        return self.api_subscription()[1].get('redirect_uri', None)

    def api_subscription(self, **kwargs):
        return self.api.subscription(
            subscription_id=self.pk,
            access_token=self.access_token, 
            callback=self.instance_update, **kwargs)

    def api_subscription_cancel(self, **kwargs):
        return self.api.subscription.cancel(
            subscription_id=self.pk,
            access_token=self.access_token, 
            callback=self.instance_update, **kwargs)

    def api_subscription_modify(self, **kwargs):
        return self.api.subscription.modify(
            subscription_id=self.pk,
            access_token=self.access_token, 
            callback=self.instance_update, **kwargs)

    def api_subscription_charge_find(self, **kwargs):
        return self.api.subscription_charge.find(
            subscription_id=self.pk,
            access_token=self.access_token, **kwargs)



class SubscriptionChargeApi(Api):

    @cached_property
    def access_token(self):
        return self.subscription_plan.access_token
    
    def api_subscription_charge(self, **kwargs):
        return self.api.subscription_charge(
            subscription_charge_id=self.pk,
            access_token=self.access_token, 
            callback=self.instance_update, **kwargs)

    def api_subscription_charge_refund(self, **kwargs):
        return self.api.subscription_charge.refund(
            subscription_charge_id=self.pk,
            access_token=self.access_token, 
            callback=self.instance_update, **kwargs)

    


from distutils.core import setup
import djwepay
import os.path

short_description = 'Django WePay Application.'
long_description = open('README.rst').read() if os.path.isfile('README.rst') \
                   else short_description

setup(
    name='django-wepay',
    version=djwepay.get_version(),
    packages=['djwepay'],
    description=short_description,
    long_description=long_description,
    author='lehins',
    author_email='lehins@yandex.ru',
    license='MIT License',
    url='https://github.com/lehins/django-wepay',
    platforms=["any"],
    install_requires=[
        'python-wepay>=1.2.0'
    ],
    dependency_links=[
        'http://github.com/lehins/python-wepay/tarball/master#egg=python-wepay-1.2.0'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
)



import cookielib, mechanize, urllib, urllib2, urlparse, random, time, warnings
from decimal import Decimal

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.management import call_command
from django.core.urlresolvers import reverse
from django.test import TestCase
from django.test.client import Client
from django.test.utils import override_settings

from djwepay.api import get_wepay_model
from djwepay.signals import *
from wepay.exceptions import WePayError

TEST_LOGIN = getattr(settings, 'WEPAY_TESTS_LOGIN', '')
TEST_PASSWORD = getattr(settings, 'WEPAY_TESTS_PASSWORD', '')
TEST_APP_ID = getattr(settings, 'WEPAY_TESTS_APP_ID')
TEST_USER_ID = getattr(settings, 'WEPAY_TESTS_USER_ID')
TEST_ACCOUNT_ID = getattr(settings, 'WEPAY_TESTS_ACCOUNT_ID')
TEST_THEME = {
    'name': 'Test Theme',
    'primary_color': '000001',
    'secondary_color': '000002',
    'background_color': '000003',
    'button_color': '000004'
}
TEST_GAQ_DOMAINS = [
    'UA-23421-01',
    'UA-23421-02'
]
TEST_CC = {
    'user_name': "Test Dude",
    'cc_number': "4003830171874018",
    'expiration_month': "02",
    'expiration_year': "17",
    'cvv': "123",
}
TEST_ADDRESS = {
    'address1': "123 Main Stret",
    'city': "Albuquerque",
    'state': "NM",
    'zip': "87108",
    'country': "US"
}

App = get_wepay_model('app')
User = get_wepay_model('user')
Account = get_wepay_model('account')
Checkout = get_wepay_model('checkout')
Preapproval = get_wepay_model('preapproval')
Withdrawal = get_wepay_model('withdrawal')
CreditCard = get_wepay_model('credit_card')



@override_settings(WEPAY_APP_ID=TEST_APP_ID)
class WePayTestCase(TestCase):
    #fixtures = ['djwepay_testdata.xml']

    @staticmethod
    def browser_create():
        browser = mechanize.Browser()
        cj = cookielib.LWPCookieJar()
        browser.set_cookiejar(cj)
        browser.set_handle_equiv(True)
        # browser.set_handle_gzip(True)
        browser.set_handle_redirect(True)
        browser.set_handle_referer(True)
        browser.set_handle_robots(False)
        browser.set_handle_refresh(
            mechanize._http.HTTPRefreshProcessor(), max_time=1)
        # debugging stuff
        #browser.set_debug_redirects(True)
        #browser.set_debug_responses(True)
        #browser.set_debug_http(True)
        browser.addheaders = [
            ('User-Agent' , 
             "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Ubuntu Chromium/28.0.1500.71 Chrome/28.0.1500.71 Safari/537.36")
        ]
        return browser

    @classmethod
    def setUpClass(cls):
        call_command('loaddata', 'djwepay_testdata')
        try:
            App.objects.get_current()
            cls.app = App.objects.get(pk=TEST_APP_ID)
            cls.user = User.objects.get(pk=TEST_USER_ID)
            cls.account = Account.objects.get(pk=TEST_ACCOUNT_ID)
        except App.DoesNotExist, e:
            raise ImproperlyConfigured(
                "There is no test data in the database. Reference the documentation. "
                "Error: %s" % e)
        if cls.app.production or not cls.app.access_token.startswith('STAGE_'):
            raise ImproperlyConfigured(
                "Set WEPAY_TESTS_APP_ID to an id of a non production App. "
                "Cannot run tests with App.production set to True.")
        cls.browser = cls.browser_create()
        cls.test_uri = reverse('wepay:tests_callback')
        cls.client = Client()


    @classmethod
    def tearDownClass(cls):
        # if any of the tests where temporary objects are created fails, 
        # tearDownClass performs the best attempt in removing them from WePay system
        (_, accs) = cls.user.api_account_find()
        acc_ids = [x['account_id'] for x in accs 
                   if x['account_id'] != TEST_ACCOUNT_ID]
        for acc_id in acc_ids:
            try:
                cls.user.api.account.delete(
                    account_id=acc_id, access_token=cls.user.access_token)
            except WePayError: pass

    def _reference_id(self):
        return random.randint(1, 10000000)

    def _validate_equality(self, obj, response):
        for key, value in response.iteritems():
            if hasattr(obj, key):
                self.assertEqual(value, getattr(obj, key))

    def test_app(self):
        response = self.app.api_app()
        self._validate_equality(*response)

    def test_app_modify(self):
        self.app.api_app()
        gaq_domains = self.app.gaq_domains
        theme_object = self.app.theme_object
        if not theme_object is None:
            theme_object.pop('theme_id', None)
        response = self.app.api_app_modify(
            theme_object=TEST_THEME, gaq_domains=TEST_GAQ_DOMAINS)
        self._validate_equality(*response)
        # modify test passed, lets change back the values
        response = self.app.api_app_modify(
            theme_object=theme_object, gaq_domains=gaq_domains)
        self._validate_equality(*response)
        
    def test_oauth2(self):
        """ Tests calls: /oauth2/authorize, /oauth2/token, /user/modify."""
        if TEST_LOGIN and TEST_PASSWORD:
            # get authentication url
            auth_url = self.app.api_oauth2_authorize(redirect_uri=self.test_uri)
            self.browser.open(auth_url)
            # fillout the form
            self.browser.select_form(nr=0)
            self.browser.form['email'] = TEST_LOGIN
            self.browser.form['password'] = TEST_PASSWORD
            self.browser.form.set_all_readonly(False)
            self.browser.form['form'] = 'login'
            try:
            # submit form and in case any Http errors occur will catch them 
            # and still extract a code from url
                self.browser.submit()
            except mechanize.HTTPError: pass
            url = self.browser.response().geturl()
            self.assertTrue(url.startswith(self.app.api.get_full_uri(self.test_uri)))
            code = urlparse.parse_qs(urlparse.urlparse(url).query)['code'][0]
            user, response = self.app.api_oauth2_token(
                redirect_uri=self.test_uri, code=code)
            user.api_user_modify(callback_uri=self.test_uri)
            self.assertEqual(TEST_LOGIN, user.email)
            self._validate_equality(user, response)
        else:
            print "Cannot test OAuth2 calls, set 'WEPAY_TESTS_LOGIN' and " \
                "'WEPAY_TESTS_PASSWORD' constants in settings."

    def test_user(self):
        """Tests calls: /user, IPNs: User, signals:ipn_received, state_changed"""
        self.user.user_name = 'foo bar'
        self.user.first_name = 'foo'
        self.user.last_name = 'bar'
        response = self.user.api_user()
        self._validate_equality(*response)
        # testing IPN and signals
        correct_state = self.user.state
        incorrect_state = 'pending' if correct_state == 'registered' else 'registered'
        signals_received = {
            'ipn_processed': False,
            'state_changed': False
        }
        def ipn_tester(instance, **kwargs):
            signals_received.update({'ipn_processed': True})
            self.assertEqual(instance.state, correct_state)
        ipn_processed.connect(ipn_tester, sender=User, weak=False)
        def state_tester(instance, previous_state, **kwargs):
            signals_received.update({'state_changed': True})
            self.assertEqual(instance.state, correct_state)
            self.assertEqual(previous_state, incorrect_state)
        state_changed.connect(state_tester, sender=User, weak=False)
        self.user.state = incorrect_state
        self.user.save()
        # imitate IPN
        response = self.client.post(
            reverse('wepay:ipn', kwargs={'obj_name':'user'}), 
            {'user_id': self.user.pk})
        self.assertEqual(response.status_code, 200)
        self.assertTrue(signals_received['ipn_processed'])
        self.assertTrue(signals_received['state_changed'])

    def test_account(self):
        """Tests calls: /account/create, /account/modify, /account/delete
        """
        name = "Runtime Test Account"
        acc, response = self.user.api_account_create(
            name=name, description="Account created during running of the "
            "django-wepay tests, it wasn't deleted probably due to a failing "
            "test case. You can safely remove it.", 
            reference_id=self._reference_id(), type='nonprofit',
            image_uri='http://www.placekitten.com/500/500',
            gaq_domains=TEST_GAQ_DOMAINS, theme_object=TEST_THEME, mcc=7299)
        self._validate_equality(acc, response)
        self.assertEqual(
            acc.api.get_full_uri(reverse(
                'wepay:ipn', kwargs={'obj_name':'account', 'user_id':self.user.pk})),
            acc.callback_uri)
        name_modified = name + " - Modified"
        acc.api_account_modify(
            name=name_modified, description="description modified", 
            reference_id=self._reference_id(), callback_uri=self.test_uri, 
            image_uri='http://www.placekitten.com/450/450',
            gaq_domains=TEST_GAQ_DOMAINS, theme_object=TEST_THEME
        )
        self.assertEqual(acc.api.get_full_uri(self.test_uri), acc.callback_uri)

        self.assertEqual(name_modified, acc.name)

        acc.api_account_delete(reason='test finished')

      
    def test_checkout(self):
        checkout, response = self.account.api_checkout_create(
            short_description="Test Checkout", type="DONATION", amount=1010, 
            long_description="Long description testing Checkout",
            payer_email_message="test@example.com", payee_email_message="Ssup",
            reference_id=self._reference_id(), app_fee=2, fee_payer='payee',
            redirect_uri=self.test_uri, auto_capture=True, require_shipping=True,
            shipping_fee=2, charge_tax=True, mode='iframe', prefill_info={
            "name": "Bill Clerico", "phone_number": "855-469-3729"},
            funding_sources='bank')
        self._validate_equality(checkout, response)
        self.assertIsNotNone(checkout.redirect_uri)
        self.assertIsNotNone(checkout.callback_uri)
        #self.assertIsNotNone(checkout.dispute_uri) - empty on new checkouts
        response = checkout.api_checkout_modify(callback_uri=self.test_uri)
        self.assertEqual(
            checkout.api.get_full_uri(self.test_uri), checkout.callback_uri)
        try:
            checkout.api_checkout_capture()
        except WePayError, e:
            self.assertTrue(str(e).rfind("4004") >= 0)
            self.assertEqual(e.code, 4004)
        try:
            checkout.api_checkout_refund(refund_reason="test refund reason")
        except WePayError, e:
            self.assertEqual(e.code, 4004)
        try:
            checkout.api_checkout_cancel(cancel_reason="test cancel reason")
        except WePayError, e:
            self.assertEqual(e.code, 4004)

    def test_finds(self):
        self.user.api_account_find(
            name='some name', reference_id=12345, sort_order='DESC')

        self.account.api_checkout_find(
            start=0, limit=10, reference_id=12345, state='new',
            preapproval_id=12345, start_time=int(time.time()-1000),
            end_time=int(time.time()), sort_order='ASC', shipping_fee=3)

        self.account.api_preapproval_find(
            state='new', reference_id=12355, start=int(time.time()-1000),
            sort_order='ASC', last_checkout_id=65865, shipping_fee=3)

        self.account.api_withdrawal_find(
            limit=5, start=int(time.time()-1000), sort_order='ASC')
            
        self.app.api_credit_card_find(reference_id=12345, limit=6, 
                                      start=int(time.time()-1000), sort_order='ASC')
        

    def test_preapproval(self):
        """Tests calls: /preapproval, /preapproval/create, /preapproval/modify,
        /preapproval/cancel
        """
        preapproval, response = self.account.api_preapproval_create(
            short_description="Testing preapprovals", period="daily", 
            amount=Decimal('1000.00'), reference_id=self._reference_id(),
            app_fee=Decimal('99.99'), fee_payer="payee_from_app",
            redirect_uri=self.test_uri,
            fallback_uri=self.test_uri, require_shipping=True,
            shipping_fee=Decimal('5.00'), charge_tax=True,
            payer_email_message="Testin preapproval message for payer",
            long_description="Very long description testing preapprovals",
            frequency=60, start_time=int(time.time()+600), 
            end_time=int(time.time()+5000000), auto_recur=True, mode="iframe",
            prefill_info={'name': "Joe Blow"}, funding_sources="bank,cc")
        self.assertIsNotNone(preapproval.preapproval_uri)
        self.assertIsNotNone(preapproval.manage_uri)
        self.assertIsNotNone(preapproval.redirect_uri)
        self.assertIsNotNone(preapproval.callback_uri)
        preapproval.api_preapproval_modify(callback_uri=self.test_uri)
        preapproval.api_preapproval_cancel()

        
    def test_withdrawal(self):
        """Tests calls: /withdrawal, /withdrawal/create, /withdrawal/modify"""
        try:
            withdrawal, response = self.account.api_withdrawal_create(
                redirect_uri=self.test_uri,
                note="Short description, testing withdrawal create.", mode="iframe")
        except WePayError, e:
            self.assertEqual(e.code, 3005)
            return
        self.assertIsNotNone(withdrawal.redirect_uri)
        self.assertIsNotNone(withdrawal.callback_uri)
        self.assertIsNotNone(withdrawal.withdrawal_uri)
        test_uri_modified = self.test_uri + "?modified=True"
        withdrawal.api_withdrawal_modify(callback_uri=test_uri_modified)
        self.assertEqual(
            withdrawal.api.get_full_uri(test_uri_modified), withdrawal.callback_uri)


    def test_credit_card(self):
        if CreditCard is None:
            return
        cc, response = self.app.api_credit_card_create(
            email='test@example.com', address=TEST_ADDRESS, **TEST_CC)
        cc.api_credit_card()
        self.assertEqual(cc.user_name, TEST_CC['user_name'])
        try:
            cc.api_credit_card_authorize()
        except WePayError, e:
            self.assertEqual(e.code, 2009)
        cc.api_credit_card_delete()

    def test_flow(self):
        ccs = [c for c in self.app.api_credit_card_find()[1] 
               if c['state'] == 'authorized']
        pas = self.account.api_preapproval_find(state="approved", limit=5)[1]
        if len(pas) > 0:
            try:
                pa = Preapproval.objects.get(pk=pas[0]['preapproval_id'])
            except Preapproval.DoesNotExist:
                pa = Preapproval.objects.create_from_response(self.account, pas[0])
        elif len(ccs) > 0:
            pa, r = self.account.api_preapproval_create(
                amount=5000, short_description="testing charge with cc app wise",
                period="daily", frequency=24, 
                payment_method_id=ccs[0]['credit_card_id'], payment_method_type='credit_card')
        try:
            co, r = self.account.api_checkout_create(
                short_description="testing flow of wepay-django", type="GOODS",
                amount=100, preapproval_id=pa.pk)
        except WePayError, e: # in case if cannot charge using this preapproval
            self.assertEqual(e.code, 1008) 

        if self.account.balances[0]['balance'] > 0:
            wd, r = self.account.api_withdrawal_create(
                redirect_uri=self.test_uri, 
                note="Short description, testing withdrawal create.", mode="iframe")
            wd.api_withdrawal_modify(callback_uri=self.test_uri)
            self.assertIsNotNone(wd.redirect_uri)
            self.assertIsNotNone(wd.callback_uri)
            self.assertIsNotNone(wd.withdrawal_uri)

    def test_batch(self):
        batch_id=1234
        self.app.api_credit_card_find(
            batch_id=batch_id, batch_mode=True, limit=2)
        self.app.api_preapproval_find(
            batch_id=batch_id, batch_mode=True, limit=3)
        response = self.app.api_batch_create(batch_id=batch_id)[1]
        self.assertEqual(len(response['calls']), 2)
